library(pls)
library(glmnet)
library(rpart)

K=10 # number of cross-validation folds
MIN_NONCONST_COLS_X = 5 # minimum number of non-constant X-data columns in PLS prediction

rm(".Random.seed")

###############################################
######## MACHINE LEARNING MODELS ##############
###############################################

# Helper function: Optimal number of components in PLS
find_ncomp = function(x) {
  comp_opt = 0
  for(comp in seq(1,length(x)-1)) {
    if(x[comp+1]>x[comp]) {
      comp_opt = comp-1
      break
    }
  }
  return(comp_opt)
}

failed_prediction = function(n, d, return_model, return_preds ) {
  print("Failed prediction due to too little variance in data" )
  predmatrix = matrix(NA, n, d)
  cors = rep.int(NA,ncol(data_y))
  varsExpl = rep.int(NA,d)
  models = rep.int(NA,d)
  ret = list()
  ret$cors = cors
  ret$varsExpl = varsExpl
  if(return_model)
    ret$models = models
  if(return_preds)
    ret$preds
}

# for testing purposes only
rundummy = function(data_x,data_y,shuffle=FALSE,seed=-1) {
  set.seed(seed)
  if(shuffle) {
    cors = runif(ncol(data_y),-1,0.3)
    varsExpl = runif(ncol(data_y),-1,0.1)
  } else {
    cors = runif(ncol(data_y),-0.1,0.9)
    varsExpl = runif(ncol(data_y),-0.1,0.45)
  }
  names(cors) = colnames(data_y)
  names(varsExpl) = colnames(data_y)
  return(list(cors=cors,varsExpl=varsExpl))
}


## PLS
runpls = function(data_x,data_y,shuffle=FALSE,seed=1, return_model=F, return_preds=F, poscomp=T) {
  set.seed(seed)
  if(shuffle) {
    data_y = data_y[sample(nrow(data_y)),,drop=F]
  }
  predmatrix = matrix(NA, nrow(data_y), ncol(data_y))
  cors = c()
  varsExpl = c()
  models = list()
  for(d in seq(1,ncol(data_y))) { # drug
    notNA = which(!is.na(data_y[,d]) & data_y[,d]!=-1)
    
    # return list of models (does not properly cross-validate number of PLS components)
    train_y = data_y[notNA,d]
    train_x = scale(data_x[notNA,,drop=FALSE])
    
    #if variance is 0 across a column, the whole column will become NA
    NAs_due_to_scaling = which(apply(train_x, 2, function(x) !any(!is.na(x))))
    if(length(NAs_due_to_scaling)>0) {
      train_x = train_x[,-NAs_due_to_scaling, drop=FALSE]
      
      # there has to be at least MIN_NONCONST_COLS_X colums with non-zero variance
      if(ncol(train_x)<=MIN_NONCONST_COLS_X) {
        return(failed_prediction(nrow(data_y),ncol(data_y), return_model, return_preds))
      }
    }
    
    if(return_model) {
      model = plsr(train_y ~ train_x,na.action = na.omit,validation="LOO", ncomp=min(ncol(train_x)-2,10),scale=FALSE)
      models[[colnames(data_y)[d]]] = model
      rmsep = RMSEP(model)$val[2,,]
      comp_opt = find_ncomp(rmsep)
      if(poscomp & comp_opt==0)
        comp_opt=1
      models[[paste(colnames(data_y)[d],"_ncomp",sep="")]] = comp_opt
    }
    
    preds = matrix(NA, nrow=1, ncol=length(notNA))
    segments = cvsegments(length(notNA), K)
    for(i in seq(1,length(segments))) { # loo, properly cross-validate number of PLS components
      train_y = data_y[notNA[-segments[[i]]],d]
      train_x = data_x[notNA[-segments[[i]]],]
      NAs_due_to_scaling = which(apply(scale(train_x), 2, function(x) !any(!is.na(x))))
      if(length(NAs_due_to_scaling)>0) {
        train_x = train_x[,-NAs_due_to_scaling, drop=F]
      }
      
      train_x = as.matrix(scale(train_x))
      test_x = data_x[notNA[segments[[i]]], colnames(train_x),drop=FALSE]
      test_x = as.matrix(scale(test_x,center=attr(train_x,"scaled:center"),scale=attr(train_x,"scaled:scale")))
   
      model = plsr(train_y ~ train_x,na.action = na.omit,validation="CV", segments=K, ncomp=min(ncol(train_x)-2,10), scale=FALSE)
      rmsep = RMSEP(model)$val[2,,] # adjustedCV
      comp_opt = find_ncomp(rmsep)
      if(poscomp & comp_opt==0)
        comp_opt = 1
      
      if(comp_opt>0) {
        preds[segments[[i]]] = predict(model,ncomp=comp_opt,newdata=test_x)
      } else {
        preds[segments[[i]]] = mean(data_y[notNA[-segments[[i]]],d])
      }
    } # cv
    
    predmatrix[notNA,d] = preds[1,]
    cors=c(cors,cor(data_y[notNA,d],preds[1,],method = "spearman", use="complete.obs"))
    varsExpl=c(varsExpl,R2(data_y[notNA,d],preds[1,]))
    
  } # drug
  colnames(predmatrix) = colnames(data_y)
  rownames(predmatrix) = rownames(data_y)
  names(cors) = colnames(data_y)
  names(varsExpl) = colnames(data_y)
  
  ret = list(cors=cors, varsExpl=varsExpl)
  
  if(return_model) {
    ret$models=models
  } 
  if(return_preds) {
    ret$preds=predmatrix
    ret$data_y = data_y
  } 
  return(ret)
}

## PLS - LATE INTEGRATION FROM TWO SOURCES
runpls_twosources = function(data_x1,data_x2,data_y,shuffle=FALSE,seed=1, poscomp=T) {
  set.seed(seed)
  if(shuffle) {
    data_y = data_y[sample(nrow(data_y)),,drop=F]
  }
  
  cors = c()
  varsExpl = c()
  for(d in seq(1,ncol(data_y))) { # drug
    notNA = which(!is.na(data_y[,d]) & data_y[,d]!=-1)
    preds = matrix(NA, nrow=1, ncol=length(notNA))
    segments = cvsegments(length(notNA), K)
    for(i in seq(1,length(segments))) {
      train_y = data_y[notNA[-segments[[i]]],d]
      train_x1 = as.matrix(scale(data_x1[notNA[-segments[[i]]],]))
      train_x2 = as.matrix(scale(data_x2[notNA[-segments[[i]]],]))
      test_x1 = as.matrix(scale(data_x1[notNA[segments[[i]]],,drop=FALSE],center=attr(train_x1,"scaled:center"),scale=attr(train_x1,"scaled:scale")))
      test_x2 = as.matrix(scale(data_x2[notNA[segments[[i]]],,drop=FALSE],center=attr(train_x2,"scaled:center"),scale=attr(train_x2,"scaled:scale")))
      
      model1 = plsr(train_y ~ train_x1,na.action = na.omit,validation="CV", segments=K, ncomp=10,scale=FALSE)
      model2 = plsr(train_y ~ train_x2,na.action = na.omit,validation="CV", segments=K, ncomp=10,scale=FALSE)
      rmsep1 = RMSEP(model1)$val[2,,] # pick adjustedCV, doesnt matter too much
      rmsep2 = RMSEP(model2)$val[2,,] # pick adjustedCV, doesnt matter too much
      comp_opt1 = find_ncomp(rmsep1)
      comp_opt2 = find_ncomp(rmsep2)
      if(poscomp & comp_opt1==0)
        comp_opt1 = 1
      if(poscomp & comp_opt2==0)
        comp_opt2 = 1
      
      rms1 = rmsep1[comp_opt1+1]
      rms2 = rmsep2[comp_opt2+1]
      
      if(comp_opt1>0) {
        preds1 = predict(model1,ncomp=comp_opt1,newdata=test_x1)
      } else {
        preds1 = mean(data_y[notNA[-i],d])
      }      
      
      if(comp_opt2>0) {
        preds2 = predict(model2,ncomp=comp_opt2,newdata=test_x2)
      } else {
        preds2 = mean(data_y[notNA[-i],d])
      }
      
      # option 1: pick best model
      #if(rms1<rms2) {
      #  preds[segments[[i]]] = preds1
      #} else {
      #  preds[segments[[i]]] = preds2
      #}
      
      # option 2: weight models according to inverse rms
      preds[segments[[i]]] = (rms2*preds1+rms1*preds2)/(rms1+rms2)
      
    } # cv
    
    cors=c(cors,cor(data_y[notNA,d],preds[1,],method = "spearman"))
    varsExpl=c(varsExpl,R2(data_y[notNA,d],preds[1,]))
    
  } # drug
  names(cors) = colnames(data_y)
  names(varsExpl) = colnames(data_y)
  
  
  return(list(cors=cors,varsExpl=varsExpl))
}

## PLS (multi-output)
runplsmulti = function(data_x,data_y,shuffle=FALSE,seed=1, poscomp=T) {
  set.seed(seed)
  if(shuffle) {
    data_y = data_y[sample(nrow(data_y)),,drop=F]
  }
  complete_cases = complete.cases(data_y) # remove rows with missing values
  data_y = data_y[complete_cases,,drop=F] 
  data_x = data_x[complete_cases,]
  # Use nested cross-validation to choose number of PLS components
  preds = matrix(NA,nrow=nrow(data_y),ncol=ncol(data_y))
  colnames(preds) = colnames(data_y)
  segments = cvsegments(nrow(data_y), K)
  for(i in seq(1,length(segments))) {
    train_y = as.matrix(data_y[-segments[[i]],,drop=F])
    train_x = data_x[-segments[[i]],]
    NAs_due_to_scaling = which(apply(scale(train_x), 2, function(x) !any(!is.na(x))))
    if(length(NAs_due_to_scaling)>0)
      train_x = train_x[,-NAs_due_to_scaling]
    
    train_x = as.matrix(scale(train_x))
    test_x = data_x[segments[[i]], colnames(train_x),drop=FALSE]
    test_x = as.matrix(scale(test_x,center=attr(train_x,"scaled:center"),scale=attr(train_x,"scaled:scale")))
    
    model = plsr(train_y ~ train_x,na.action = na.omit,validation="CV", segments=K, ncomp=10,scale=FALSE)
    rmsep = RMSEP(model)$val[2,,] # pick adjustedCV, doesnt matter too much
    rmsep_tot = apply(rmsep*rmsep,2,sum) # add in quadratures
    comp_opt = find_ncomp(rmsep_tot)
    if(poscomp & comp_opt==0)
      comp_opt = 1
    if(comp_opt>0) {
      preds[segments[[i]],] = predict(model,ncomp=comp_opt,newdata=test_x)
    } else {
      preds[segments[[i]],] = apply(data_y[-i,,drop=F],2,mean)
    }
  }
  
  cors = c()
  for(drug in colnames(data_y)) {
    cors=c(cors,cor(data_y[,drug],preds[,drug],method = "spearman"))
  }
  names(cors) = colnames(data_y)
  
  varsExpl = c()
  for(drug in colnames(data_y)) {
    varsExpl = c(varsExpl,R2(data_y[,drug],preds[,drug]))
  }
  names(varsExpl) = colnames(data_y)  
  
  return(list(cors=cors,varsExpl=varsExpl))
}

## ELASTIC NET
runelasticnet = function(data_x,data_y,alphas=seq(0,1,by=0.1),shuffle=FALSE,seed=1) {
  # Use nested cross-validation to choose number of PLS components
  set.seed(seed)
  if(shuffle) {
    data_y = data_y[sample(nrow(data_y)),,drop=F]
  } else if(seed != -1) {
    set.seed(seed)
  }
  
  cors = c()
  varsExpl = c()
  for(d in seq(1,ncol(data_y))) { # drug
    notNA = which(!is.na(data_y[,d]) & data_y[,d]!=-1)
    preds = matrix(NA, nrow=1, ncol=length(notNA))
    segments = cvsegments(length(notNA), K)
    for(i in seq(1,length(segments))) {
      train_y = data_y[notNA[-segments[[i]]],d]
      train_x = data_x[notNA[-segments[[i]]],]
      NAs_due_to_scaling = which(apply(scale(train_x), 2, function(x) !any(!is.na(x))))
      if(length(NAs_due_to_scaling)>0)
        train_x = train_x[,-NAs_due_to_scaling]
      
      train_x = as.matrix(scale(train_x))
      test_x = data_x[notNA[segments[[i]]], colnames(train_x),drop=FALSE]
      test_x = as.matrix(scale(test_x,center=attr(train_x,"scaled:center"),scale=attr(train_x,"scaled:scale")))
      
      rms_min = 99999999999999
      model_best = 0
      for(a in alphas) {   # alpha
        model =  cv.glmnet(train_x, train_y, alpha=a, nfolds=K, maxit = 10000)
        
        rms = model$cvm[which(model$lambda==model$lambda.min)]
        if(rms<rms_min) {
          model_best = model
        }
      } # alpha
      preds[segments[[i]]] = predict(model_best,newx = test_x,s = "lambda.min")
    } # loo
    cors=c(cors,cor(data_y[notNA,d],preds[1,],method = "spearman"))
    varsExpl = c(varsExpl,R2(data_y[notNA,d],preds[1,]))
    
  } # drug
  names(cors) = colnames(data_y)
  names(varsExpl) = colnames(data_y)
  
  return(list(cors=cors,varsExpl=varsExpl)) 
}

## REGRESSION TREE
runregressiontree = function(data_x,data_y,shuffle=FALSE,seed=1) {
  set.seed(seed)
  if(shuffle) {
    data_y = data_y[sample(nrow(data_y)),,drop=F]
  }
  
  cors = c()
  varsExpl = c()
  for(d in seq(1,ncol(data_y))) { # drug
    notNA = which(!is.na(data_y[,d]) & data_y[,d]!=-1)
    preds = matrix(NA, nrow=1, ncol=length(notNA))
    segments = cvsegments(length(notNA), K)
    for(i in seq(1,length(segments))) { 
      train_y = data_y[notNA[-segments[[i]]],d]
      train_x = data_x[notNA[-segments[[i]]],]
      NAs_due_to_scaling = which(apply(scale(train_x), 2, function(x) !any(!is.na(x))))
      if(length(NAs_due_to_scaling)>0)
        train_x = train_x[,-NAs_due_to_scaling]
      
      train_x = as.matrix(scale(train_x))
      test_x = data_x[notNA[segments[[i]]], colnames(train_x),drop=FALSE]
      test_x = as.matrix(scale(test_x,center=attr(train_x,"scaled:center"),scale=attr(train_x,"scaled:scale")))
      
      model = rpart(train_y ~ . , data=data.frame(cbind(train_y,train_x)), method="anova", control = rpart.control(minbucket=5, xval = K))
      optcp <- model$cptable[which.min(model$cptable[,"xerror"]),"CP"]
      model_pruned = prune(model, cp = optcp)
      
      preds[segments[[i]]] = predict(model_pruned,newdata=data.frame(test_x),type="vector")
      
    }
    
    cors=c(cors,cor(data_y[notNA,d],preds[1,],method = "spearman"))
    varsExpl=c(varsExpl,R2(data_y[notNA,d],preds[1,]))
    
  } # drug
  names(cors) = colnames(data_y)
  names(varsExpl) = colnames(data_y)
  
  
  return(list(cors=cors,varsExpl=varsExpl))
}

## TISSUE ORIGIN
runtissue = function(data_x,data_y,shuffle=FALSE,seed=1) {
  set.seed(seed) # not used for LOO-CV
  if(shuffle) {
    data_y = data_y[sample(nrow(data_y)),,drop=F]
    rownames(data_y) = rownames(data_x) # need to do this, to make sen/tissue join work
  }
  
  if(!('tissue' %in% colnames(data_x))) {
    print("MISSING TISSUE INFORMATION")
    return(NULL)
  }

  cors = c()
  varsExpl = c()
  for(d in seq(1,ncol(data_y))) { # drug
    notNA = which(!is.na(data_y[,d]) & data_y[,d]!=-1) # -1 is legacy NA (matlab)
    
    notNA_tissueTable = table(data_x$tissue[notNA])
    notNA_noTissues = names(notNA_tissueTable[notNA_tissueTable<=1])
    notNA = setdiff(notNA, which(data_x$tissue %in% notNA_noTissues) )
    preds = c()
    for(i in seq(1,length(notNA))) { # loo
      train_y = data_y[notNA[-i], d, drop=F]
      train_x = data_x[notNA[-i],, drop=F]
      
      test_x = data_x[notNA[i],,drop=FALSE]
      model = inner_join(train_x %>% mutate(Cell.Line=rownames(train_x)),
                         train_y %>% mutate(Cell.Line=rownames(train_y)), by="Cell.Line")
      
      model_back = model
      
      model = model %>% 
        dplyr::select(-Cell.Line) %>% 
        group_by(tissue) %>% 
        summarise_all(funs(mean(., na.rm = TRUE))) %>% 
        dplyr::select(tissue, everything())
      
      pred = model[model$tissue==test_x$tissue, 2][[1]]
      preds = c(preds, pred)
    }

    cors=c(cors,cor(data_y[notNA,d],preds,method = "spearman"))
    varsExpl=c(varsExpl,R2(data_y[notNA,d],preds))
  }
  names(cors) = colnames(data_y)
  names(varsExpl) = colnames(data_y)
  
  return(list(cors=cors,varsExpl=varsExpl))
}


## PLS
runmaxcor = function(data_x,data_y,shuffle=FALSE,seed=1, return_model=F, return_preds=F) {

    maxcormodel = function(X, y) {
    spearman <- rep(NA, ncol(X))
    names(spearman) = colnames(X)
    for(x in colnames(X)) {
        spearman[x] = cor(X[,x], y, method = "spearman", use="complete.obs")
    }
    maxcorvar = names(spearman)[which.max(abs(spearman))]  
    df = data.frame(x=X[,maxcorvar], y=y)
    model=lm(formula = y ~ x, data = df)

    return(list(model=model, maxcorvar=maxcorvar))
  }
  
    if(shuffle) {
    data_y = data_y[sample(nrow(data_y)),,drop=F]
  }
  set.seed(seed)
  
  predmatrix = matrix(NA, nrow(data_y), ncol(data_y))
  cors = c()
  varsExpl = c()
  models = list()
  for(d in seq(1,ncol(data_y))) { # drug
    notNA = which(!is.na(data_y[,d]) & data_y[,d]!=-1)
    
    # return list of models (does not properly cross-validate number of PLS components)
    train_y = data_y[notNA,d]
    train_x = scale(data_x[notNA,,drop=FALSE])
    
    #if variance is very low across a column, the whole column might become NA
    NAs_due_to_scaling = which(apply(train_x, 2, function(x) !any(!is.na(x))))
    if(length(NAs_due_to_scaling)>0)
      train_x = train_x[,-NAs_due_to_scaling]
    
    if(return_model) {
      model = maxcormodel(train_x, train_y)
      models[[colnames(data_y)[d]]] = model
    }
    
    preds = matrix(NA, nrow=1, ncol=length(notNA))
    segments = cvsegments(length(notNA), K)
    for(i in seq(1,length(segments))) { # loo, properly cross-validate number of PLS components
      train_y = data_y[notNA[-segments[[i]]],d]
      train_x = data_x[notNA[-segments[[i]]],]
      NAs_due_to_scaling = which(apply(scale(train_x), 2, function(x) !any(!is.na(x))))
      if(length(NAs_due_to_scaling)>0)
        train_x = train_x[,-NAs_due_to_scaling]
      
      train_x = as.matrix(scale(train_x))
      test_x = data_x[notNA[segments[[i]]], colnames(train_x),drop=FALSE]
      test_x = as.matrix(scale(test_x,center=attr(train_x,"scaled:center"),scale=attr(train_x,"scaled:scale")))
      
      model = maxcormodel(train_x, train_y)
      
      preds[segments[[i]]] = predict(model$model,newdata=data.frame(x=test_x[,model$maxcorvar]))

    } # cv
    
    predmatrix[notNA,d] = preds[1,]
    cors=c(cors,cor(data_y[notNA,d],preds[1,],method = "spearman"))
    varsExpl=c(varsExpl,R2(data_y[notNA,d],preds[1,]))
    
  } # drug
  colnames(predmatrix) = colnames(data_y)
  rownames(predmatrix) = rownames(data_y)
  names(cors) = colnames(data_y)
  names(varsExpl) = colnames(data_y)
  
  ret = list(cors=cors, varsExpl=varsExpl)
  
  if(return_model) {
    ret$models=models
  } 
  if(return_preds) {
    ret$preds=predmatrix
    ret$data_y = data_y
  } 
  return(ret)
}
